import os
import polars as pl
from scipy.spatial.distance import cosine, euclidean

from local_vector_search.misc import robust_detect, pickle_save
from local_vector_search.text_cleaning import chunk_text, clean_text, yield_docs


def embed_docs(
    metadata,
    files_path,
    filepath_col_name,
    tokenizer_name="meta-llama/Llama-2-7b-hf",
    chunk_size=700,
    chunk_overlap=150,
    quiet=True,
    write_path=None,
    model_path=None,
    model=None,
    include_metadata=False,
    text_ids=None,
):
    # chunking
    counter = 0

    if text_ids is None:
        files = [_ for _ in os.listdir(files_path) if ".txt" in _]
    else:
        files = metadata.filter(pl.col("text_id").is_in(text_ids))["filepath"].to_list()

    languages = []
    for file_name in files:
        if not (quiet):
            print(f"Chunking and embedding doc {counter+1}/{len(files)}")

        with open(f"{files_path}{file_name}", "r") as file:
            s = file.read()
            languages.append(robust_detect(s))

            doc_metadata = metadata.filter(pl.col(filepath_col_name) == file_name).drop(
                filepath_col_name,
                "vector_weight",
            )

            doc_weight = (
                metadata.filter(pl.col(filepath_col_name) == file_name)
                .select("vector_weight")
                .item()
            )

            if len(doc_metadata.columns) > 0:
                metadata_string = " | ".join(
                    f"{col}: {val}"
                    for col, val in zip(doc_metadata.columns, doc_metadata.row(0))
                    if col not in ["text_id", "chunk_id"]
                )
            else:
                metadata_string = ""

            # chunk the document
            chunks, page_nums = chunk_text(
                s,
                tokenizer_name,
                chunk_size,
                chunk_overlap,
                include_metadata,
                metadata_string,
            )

            df = pl.DataFrame({"chunk": chunks, "page_num": page_nums})

            # adding page numbers to metadata
            df = df.with_columns(
                pl.col("page_num")
                .map_elements(
                    lambda x: f"{metadata_string} | page number(s): {x}",
                    return_dtype=pl.String,
                )
                .alias("metadata_string")
            )

            for col in doc_metadata.columns:
                df = df.with_columns(pl.lit(doc_metadata.select(col).item()).alias(col))

            # adding weight
            df = df.with_columns(pl.lit(doc_weight).alias("vector_weight"))

            df = df.select(
                doc_metadata.columns + ["metadata_string", "chunk", "vector_weight"]
            )

            if str(type(model)) != "<class 'gensim.models.doc2vec.Doc2Vec'>":
                df = df.with_columns(
                    pl.Series("embedding", model.encode(df.to_pandas()["chunk"]))
                )

            if counter == 0:
                final_df = df
            else:
                final_df = final_df.vstack(df)

            counter += 1

    # doc2vec
    if str(type(model)) == "<class 'gensim.models.doc2vec.Doc2Vec'>":
        corpus = [x for x in yield_docs(final_df)]
        model.build_vocab(corpus)
        model.train(corpus, total_examples=model.corpus_count, epochs=10)

        # adding embeddings to final_df
        final_df = final_df.with_columns(pl.Series("embedding", model.dv.vectors))

        if model_path is not None:
            pickle_save(model, model_path)

    # adding chunk ids
    final_df = final_df.with_columns(
        pl.int_range(0, final_df.height, dtype=pl.Int64).alias("chunk_id")
    )
    final_df = final_df.select(
        ["chunk_id"] + [col for col in final_df.columns if col != "chunk_id"]
    )

    # writing out parquet file
    if write_path is not None:
        final_df.write_parquet(write_path)

    # getting the most common corpus language
    corpus_language = max(set(languages), key=languages.count)

    return final_df, corpus_language


def get_top_n(
    query,
    final_df,
    text_ids=[],
    clean_text_function=None,
    model=None,
    top_n=3,
    distance_metric="cosine",
    chunk_text_format="Excerpt metadata: {}\n\nExcerpt: {}\n\n\n\n",
    include_metadata=False,
):
    "return the top chunks based on distance"

    if distance_metric == "cosine":
        dist_func = cosine
    elif distance_metric == "euclidean":
        dist_func = euclidean

    if clean_text_function is None:
        clean_text_function = clean_text

    transformed_query = clean_text_function(query)

    # filter for desired text ids
    if len(text_ids) > 0:
        final_df = final_df.filter(pl.col("text_id").is_in(text_ids))

    if top_n is None:
        top_n = len(final_df)

    # docv2vec
    try:
        new_vector = model.infer_vector(transformed_query.split())
    # transformer
    except:
        new_vector = model.encode(transformed_query)

    # calculating distance
    return_df = (
        final_df.with_columns(
            (
                pl.col("embedding").map_elements(
                    lambda x: dist_func(x, new_vector), return_dtype=pl.Float64
                )
                * (1 / pl.col("vector_weight"))
            ).alias("vector_distance")
        )
        .sort(by="vector_distance", descending=False)
        .limit(top_n)
    )

    chunk_ids = return_df["chunk_id"].to_list()

    # does the text format have a |, in which case this comes first
    if "|" in chunk_text_format:
        return_string = chunk_text_format.split("|")[0]
        chunk_text_format = chunk_text_format.split("|")[1]
    else:
        return_string = ""

    for row in return_df.iter_rows():
        row_dict = dict(zip(return_df.columns, row))

        if (
            include_metadata
        ):  # don't include in this case because metadata is in the chunk itself
            metadata_string = ""
        else:
            metadata_string = row_dict["metadata_string"]

        return_string += chunk_text_format.format(metadata_string, row_dict["chunk"])

    return {
        "response": return_string,
        "chunk_ids": chunk_ids,
    }


def retrieve_chunks(embeddings_df, chunk_ids):
    "retrieve text of chunks given a list of chunk_ids"
    filtered_df = embeddings_df.filter(pl.col("chunk_id").is_in(chunk_ids))
    filtered_df = (
        filtered_df.with_columns(
            pl.col("chunk_id")
            .map_elements(
                lambda x: {v: i for i, v in enumerate(chunk_ids)}.get(x, len(chunk_ids))
            )
            .alias("sort_key")
        )
        .sort("sort_key")
        .drop("sort_key")
    )

    return {
        "metadata": filtered_df["metadata_string"].to_list(),
        "chunks": filtered_df["chunk"].to_list(),
    }



from importlib import import_module
from mtranslate import translate
import polars as pl
from sentence_transformers import SentenceTransformer

from local_vector_search.misc import robust_detect


class local_vs:
    """Primary class of the library, manage and embed the corpus
    parameters:
        :metadata_path: str: path to the metadata.csv file. Must have at least the column "filepath" corresponding to the name of the files in the 'files_path' directory. Include a column named 'vector_weight' in the metadata csv to weight distances. 1 = no alteration to the distance, 0.5 = distance will be multiplied by 1/0.5, so be made 2x farther. Not including this column will result in equal weight for all documents.
        :files_path: str: folder path containing the .txt files of the documents
        :filepath_col_name: str: name of the column in the metadata that contains the file names
        :model: gensim.models.doc2vec.Doc2Vec or str: if using a doc2vec model, the model with its hyperparameters, if using a pre-trained embedding model, its name
        :tokenizer_name: str: name of the tokenizer, options: https://huggingface.co/docs/transformers/model_doc/auto
        :clean_text_function: function: function that takes a single input string and returns an output string. Text will go through this process before being passed as a query for the vector similarity search
        :embeddings_path: str: if already generated the embeddings, the path to the parquet file where they are saved.
        :doc2vec_path: str: if already generated the embeddings, the path to the doc2vec pickle model
        :include_metadata: bool: whether nor not to include the metadata in the chunk so it will be searched in the vector search
    """

    def __init__(
        self,
        metadata_path=None,
        files_path=None,
        filepath_col_name="filepath",
        model="all-MiniLM-L6-v2",
        tokenizer_name="meta-llama/Llama-2-7b-hf",
        clean_text_function=None,
        embeddings_path=None,
        doc2vec_path=None,
        include_metadata=False,
    ):
        self.embed = import_module("local_vector_search.embed")
        self.misc = import_module("local_vector_search.misc")
        self.text_cleaning = import_module("local_vector_search.text_cleaning")

        self.metadata_path = metadata_path
        metadata = pl.read_csv(metadata_path)

        # add a text_id column if it's not already there
        if "text_id" not in metadata.columns:
            metadata = metadata.with_row_index(name="text_id", offset=1)

        # add a weight column if it's not already there
        if "vector_weight" not in metadata.columns:
            metadata = metadata.with_columns(pl.lit(1.0).alias("vector_weight"))

        metadata.write_csv(metadata_path)
        self.metadata = metadata

        self.files_path = files_path
        self.filepath_col_name = filepath_col_name

        if str(type(model)) != "<class 'gensim.models.doc2vec.Doc2Vec'>":
            self.model = SentenceTransformer(model)
        else:
            self.model = model

        self.tokenizer_name = tokenizer_name
        self.clean_text_function = clean_text_function
        self.embeddings_path = embeddings_path
        self.doc2vec_path = doc2vec_path
        self.include_metadata = include_metadata

        if embeddings_path is not None:
            self.embeddings_df = pl.read_parquet(embeddings_path)
            # finding corpus language
            languages = []
            for row in self.embeddings_df.sample(
                n=min(100, len(self.embeddings_df)), shuffle=True
            ).iter_rows():  # only do max 100 rows of the embeddings df to save time
                row_dict = dict(zip(self.embeddings_df.columns, row))
                languages.append(robust_detect(row_dict["chunk"]))
            try:
                self.corpus_language = max(set(languages), key=languages.count)
            except:
                self.corpus_language = "en"

    def embed_docs(
        self,
        chunk_size=700,
        chunk_overlap=150,
        embeddings_path=None,
        model_path=None,
        text_ids=None,
        quiet=True,
    ):
        """Chunk and embed the documents
        parameters:
            :chunk_size: int: how many tokens each chunk should be
            :chunk_overlap: int: how much overlap each chunk should have
            :write_path: str: where to write out the parquet file that will contain the embeddings
            :model_path: str: if using doc2vec, where to write the doc2vec model out
            :text_ids: list[int]: list of text ids in case want to embed in a loop for status printing
            :quiet: bool: whether or not to print out the embedding progress
        """

        final_df, corpus_language = self.embed.embed_docs(
            metadata=self.metadata,
            files_path=self.files_path,
            filepath_col_name=self.filepath_col_name,
            tokenizer_name=self.tokenizer_name,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            quiet=quiet,
            write_path=embeddings_path,
            model_path=model_path,
            model=self.model,
            include_metadata=self.include_metadata,
            text_ids=text_ids,
        )

        self.embeddings_path = embeddings_path
        try:
            self.embeddings_df = pl.read_parquet(embeddings_path)
        except:
            pass
        self.corpus_language = corpus_language

        if model_path is not None:
            self.model = self.misc.pickle_load(model_path)

        if embeddings_path is None:
            return final_df

    def get_top_n(
        self,
        query,
        text_ids=[],
        top_n=3,
        distance_metric="cosine",
        chunk_text_format="Here is the context information:\n\n|Excerpt metadata: '{}'\n\nExcerpt: '{}'\n\n\n\n",
    ):
        """Retrieve top n chunks according to a query
        parameters:
            :query: str: the new query
            :text_ids: list[int]: list of text ids to include in the search. Leave an empty list to search all
            :top_n: int: top n chunks to retrieve
            :distance_metric: str: "cosine" or "euclidean"
            :chunk_text_format: str: how to format the retrieved chunks, two {}'s, first will insert the metadata, second will insert the chunk. Anything you put in frot of a '|' will only appear in the beginning of the retrieval, after tha will appear for every chunk
        returns:
            :dict: dictionary, 'response' = text of chunks, 'chunk_ids' = list of top n closesrtt chunk ids
        """

        query_lang = robust_detect(query)

        if query_lang != self.corpus_language:
            query = translate(query, self.corpus_language, query_lang)

        response = self.embed.get_top_n(
            query=query,
            final_df=self.embeddings_df,
            text_ids=text_ids,
            clean_text_function=self.clean_text_function,
            model=self.model,
            top_n=top_n,
            distance_metric=distance_metric,
            chunk_text_format=chunk_text_format,
            include_metadata=self.include_metadata,
        )

        return response

    def retrieve_chunks(self, chunk_ids):
        "Retrieve metadata strings and chunks given chunk ids"
        return self.embed.retrieve_chunks(self.embeddings_df, chunk_ids)
